(function (Drupal, drupalSettings, $) {
  Drupal.behaviors.stripeConfirm = {
    attach(context) {
      const cardMethod = document.querySelector('input[name="method"][value="card"]');
      // Query inside context, in case of AJAX
      const form = context.querySelector ? context.querySelector('form.give-donation-form') : document.querySelector('form.give-donation-form');
      try {
        stripe = Stripe(drupalSettings.stripe.publishable_key);
      }
      catch (e) {
        $('#edit-method-card').attr('disabled', true);
        $('label[for="edit-method-card"]').append('<div class="error messages--error form--inline-feedback form--inline-feedback--error visible">' + e.message + '</div>');
        return;
      }

      const elements = stripe.elements();
      const card = elements.create('card', {
        hidePostalCode: true,
        style: {
          base: {
            iconColor: 'red',
            color: 'green',
            fontWeight: 400,
            fontFamily: '"Helvetica Neue", "Helvetica", sans-serif',
            fontSize: '15px',
            '::placeholder': {
              color: '#ccc',
            }
          }
        }
      });
      card.mount('#stripe-card-element');
      
      // Stripe's PaymentIntent API is a server-side API only. That means:
      // Modifying a PaymentIntent (e.g., updating the amount) requires your secret key.
      // Your secret key must never be exposed in JavaScript — it's like a master password.
      // Stripe.js (client-side) is only allowed to confirm or handle a PaymentIntent using the client_secret, not to create or modify it.
      // So we have to call back to the server to modify the amount, and then set the new client-secret
      var amountInput = document.querySelector('[data-drupal-selector="edit-amount"]');
      
      cardMethod.addEventListener('change', (event) => {
        if (event.target.checked) {
          updatePaymentIntent(drupalSettings, amountInput.value)
        }
      });

      amountInput.addEventListener('change', async function () {
        updatePaymentIntent(drupalSettings, amountInput.value);
      });
      
      // Tell drupal to tell stripe to change the amount.
      async function updatePaymentIntent(drupalSettings, newAmount) {
        if (drupalSettings.give.amount != newAmount && document.querySelector('input[name="method"]:checked').value == 'card') {
          let response = await fetch('/give/stripe/update-intent', {
            method: 'POST',
            headers: {
              'Content-Type': 'application/json',
              'X-CSRF-Token': drupalSettings.csrfToken
            },
            body: JSON.stringify({ amount: newAmount * 100, intent_id: drupalSettings.stripe.intent_id})
          });
          // check for a valid response
          if (!response.ok) {  // Checks for HTTP status 200-299
            alert(`Error updating PaymentIntent: ${response.status} ${response.statusText}`);
          }
        }
      }
      
      // Send the payment confirmation directly to stripe and then onSuccess submit the form back to Drupal.
      form.addEventListener('submit', async function (e) {
        e.preventDefault();
        try {
          const result = await stripe.confirmCardPayment(drupalSettings.stripe.client_secret, {
            payment_method: {
              card: card,
              billing_details: {
                name: drupalSettings.stripe.username,
              }
            }
          });

          if (result.error) {
            alert(result.error.message);
          }
          else if (result.paymentIntent.status === 'succeeded') {
            // Successful payment, submit the form normally to Drupal
            form.submit();
          }
        }
        catch (error) {
          alert('Payment failed: ' + error.message);
        }
      });
    }
  };
})(Drupal, drupalSettings, jQuery);
