<?php

namespace Drupal\give\Form\Donation;

use Drupal\give\Plugin\Field\FieldType\PaymentMethod;
use Drupal\give\MailHandlerInterface;
use Drupal\give\GiveStripeInterface;
use Drupal\user\Entity\User;
use Drupal\Core\Url;
use Drupal\Core\Render\Markup;
use Drupal\Core\Entity\ContentEntityForm;
use Drupal\Core\Entity\EntityRepositoryInterface;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Utility\Token;
use Drupal\Core\Extension\ModuleExtensionList;
use Drupal\Component\Datetime\TimeInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Stripe\PaymentIntent;
use Symfony\Component\HttpFoundation\RedirectResponse;

/**
 * Form controller for users to donate.
 */
class PaymentForm extends ContentEntityForm {

  const STRIPE_DEV_CODES = "Stripe is in test mode, use one of the following card numbers with any future date:
  <br />4000000000003220 Visa 3DS !important
  <br />4242424242424242 Visa + 3 digits
  <br />4000056655665556 Visa (debit) + 3 digits
  <br />5555555555554444 Mastercard + 3 digits
  <br />2223003122003222 Mastercard (2-series) + 3 digits
  <br />5200828282828210 Mastercard (debit) + 3 digits
  <br />5105105105105100 Mastercard (prepaid) + 3 digits
  <br />378282246310005 American Express + 4 digits
  <br />6011111111111117 Discover + 3 digits
  <br />3056930009020004 Diners Club + 3 digits
  <br />3566002020360505 JCB + 3 digits
  <br />6200000000000005 UnionPay";

  /**
   * The give mailHandler service.
   *
   * @var \Drupal\give\MailHandlerInterface
   */
  protected $mailHandler;

  /**
   * The Stripe Service.
   *
   * @var \Drupal\give\GiveStripeInterface
   */
  protected $giveStripe;


  /**
   * @var Drupal\Core\Utility\Token
   */
  protected $token;

  /**
   * @var ModuleExtensionList
   */
  protected $extensionListModule;

  /**
   * @var PaymentIntent
   */
  protected $paymentIntent;

  /**
   * Constructs a PaymentForm object.
   *
   * @param EntityRepositoryInterface $entity_repository
   * @param EntityTypeBundleInfoInterface $entity_type_bundle_info
   * @param TimeInterface $time
   * @param Token $token
   * @param type $extension_list_module
   * @param MailHandlerInterface $mail_handler
   * @param GiveStripeInterface $give_stripe
   */

  public function __construct($entity_repository, $entity_type_bundle_info, $time, Token $token, ModuleExtensionList $extension_list_module, MailHandlerInterface $mail_handler, GiveStripeInterface $give_stripe) {
    parent::__construct($entity_repository, $entity_type_bundle_info, $time);
    $this->mailHandler = $mail_handler;
    $this->giveStripe = $give_stripe;
    $this->token = $token;
    $this->extensionListModule = $extension_list_module;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity.repository'),
      $container->get('entity_type.bundle.info'),
      $container->get('datetime.time'),
      $container->get('token'),
      $container->get('extension.list.module'),
      $container->get('give.mail_handler'),
      $container->get('give.stripe')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    $donation = $this->entity;

    if ($donation->isCompleted()) {
      $donate_path = Url::fromRoute(
        'entity.give_form.canonical',
        ['give_form' => $donation->getGiveForm()->id()]
      )->toString();
      $form_state->setErrorByName(
        'stripe_errors',
        $this->t(
          'You have already completed this donation. Thank you! Please <a href=":donate_path">donate again</a> if you wish to give more.',
          [':donate_path' => $donate_path]
        )
      );
      // @todo deny access to payment forms for completed donations.
      $response = new RedirectResponse($donation->toUrl()->toString());
      $response->send();
    }


    $give_settings = $this->config('give.settings');
    $give_form = $donation->getGiveForm();

    $form = parent::form($form, $form_state, $donation);
    $form['#prefix'] = '<div class="flow">';
    $form['#suffix'] = '</div>';
    $form['#attributes']['class'][] = 'give-form give-form-payment flow-middle';
    $form['name']['#access'] = FALSE;
    $form['#title'] = $give_form->getPaymentPageTitle($donation);

    $form['stripe_errors'] = [
      '#markup' => '<span class="payment-errors"></span>',
      '#weight' => 1,
    ];

    if ($donation->amount->isEmpty()) {
      $form['amount'] = [
        '#title' => $this->t('Amount'),
        '#field_prefix' => STRIPE_CURRENCY_SYMBOLS[$this->config('give.settings')->get('currency_symbol')],
        '#type' => 'number',
        '#step' => 0.01,
        '#default_value' => 0,
        '#weight' => 0,
      ];
    }
    else {
      // if this is altered (by javascript) other javascript will call /give/stripe/update-intent
      $form['amount'] = [
        '#type' => 'hidden',
        '#default_value' => $donation->amount->value,
        '#input' => TRUE, // 🔑 This tells Drupal to process user-submitted value
      ];
      $form['show_amount'] = [
        '#type' => 'item',
        '#title' => $this->t('Amount'),
        '#value' => format_stripe_currency($donation->amount->value ?? 0),
        '#plain_text' => $donation->isRecurring() ?
          $this->t(':plan', [':plan' => $donation->getRecurringProductName()]) :
          $donation->getFormattedAmount(),
        '#weight' => 0,
      ];
    }

    // Tell stripe we may want to make a payment, and get a unique id back.
    $stripe_amount = intval($donation->amount->value * 100);
    $this->createPaymentIntent(
      $this->config('give.settings')->get('currency_symbol'),
      $stripe_amount,
      $donation->label()
    );

    $form['#attached'] = [
      'library' => ['give/stripe'],
      'drupalSettings' => [
        'give' => [
          'amount' => $donation->amount->value
        ],
        'stripe' => [
          // js will validate and submit the payment details when submitting this form.
          'publishable_key' => $give_settings->get('stripe_publishable_key'),
          'client_secret' => $this->paymentIntent->client_secret,
          'intent_id' => $this->paymentIntent->id,
          'username' => $donation->getOwner()->getDisplayName(),
        ],
        'http_header' => [
          ['Content-Security-Policy' => "connect-src 'https://api.stripe.com'"],
          ['Content-Security-Policy' => "child-src 'https://js.stripe.com'"],
          ['Content-Security-Policy' => "script-src 'https://js.stripe.com'"],
        ],
      ],
    ];

    $form['method'] = [
      '#type' => 'radios',
      '#title' => $this->t('Choose donation method'),
      '#options' => payment_method_names(),
      '#required' => TRUE,
      '#weight' => 1,
    ];

    $form['stripe_details'] = [
      '#type' => 'details',
      '#tree' => FALSE,
      '#open' => TRUE,
      '#weight' => 3,
      '#states' => [
        'visible' => [
          ':input[name="method"]' => ['value' => PaymentMethod::GIVE_VIA_STRIPE],
        ]
      ],
      'help_card' => [
        '#type' => 'markup',
        '#input' => FALSE,
        '#markup' => Markup::create($give_form->getCreditCardExtraText()),
        '#weight' => 1,
        '#states' => [
          'visible' => [
            ':input[name="method"]' => ['value' => PaymentMethod::GIVE_VIA_STRIPE],
          ]
        ]
      ],
      // Custom radar rules can't use name but Stripe's risk assessment does.
      // Therefore this should default to the entered name but be editable.
      // Matslats says: What is Radar? Are we even using it?
      // Matslats says: Why should data entered in stage 1 of the form be editable?
      'donor_name' => [
        '#type' => 'textfield',
        '#title' => $this->t('Name on bank account'),
        '#description' => $give_form->getDonorNameDescription(),
        '#default_value' => $donation->getDonorName(),
        '#disabled' => TRUE,
        '#weight' => 3
      ],
      'card' => [
        '#type' => 'item',
        '#title' => $this->t('Credit or debit card'),
        //'#required' => TRUE,
        //'#value' => TRUE,
        '#markup' => '<div id="stripe-card-element" class="give-card-element"></div>'
        . '<div class="form--inline-feedback form--inline-feedback--success" id="stripe-card-errors"></div>'
        . '<div class="form--inline-feedback form--inline-feedback--error" id="stripe-card-success"></div>',
        '#allowed_tags' => ['div'],
        '#weight' => 6,
      ]
    ];

    if (isset($form['method']['#options'][PaymentMethod::GIVE_VIA_STRIPE])) {
      if (substr($give_settings->get('stripe_publishable_key'), 0, 7) == 'pk_test') {
        $form['stripe_details']['dev_mode'] = [
          '#type' => 'markup',
          '#markup' => Markup::create(SELF::STRIPE_DEV_CODES),
          '#weight' => 2,
        ];
      }
    }


    $form['check_details'] = [
      '#type' => 'details',
      '#tree' => FALSE,
      '#open' => TRUE,
      '#weight' => 3,
      '#states' => [
        'visible' => [
          ':input[name="method"]' => ['value' => PaymentMethod::GIVE_VIA_CHECK],
        ],
      ],
      'help_check' => [
        '#type' => 'markup',
        '#input' => FALSE,
        '#markup' => Markup::create($give_form->getCheckOrOtherText()),
        '#maxlength' => 2000,
        '#weight' => 1
      ],
      'telephone' => [
        '#type' => 'tel',
        '#title' => $this->t('Telephone number'),
        '#maxlength' => 15,
        '#weight' => 2
      ],
      'other_info' => [
        '#type' => 'textarea',
        '#title' => $this->t('Further information'),
        '#description' => $this->t('Please ask any questions or explain anything needed to arrange for giving your donation.'),
        '#weight' => 3
      ]
    ];

    $form['bank_details'] = [
      '#type' => 'details',
      '#open' => TRUE,
      '#tree' => FALSE,
      '#weight' => 3,
      '#states' => [
        'visible' => [
          ':input[name="method"]' => ['value' => PaymentMethod::GIVE_VIA_BANK],
        ],
      ],
      'help_banktransfer' => [
        '#type' => 'markup',
        '#input' => FALSE,
        '#markup' => Markup::create($give_form->get('bankTransferDetails')),
        '#weight' => 1,
      ]
    ];


    if ($give_form->getCollectAddress()) {
      $form['address_line1'] = [
        '#type' => 'textfield',
        '#title' => $this->t('Billing address'),
        '#required' => TRUE,
        '#weight' => 4,
      ];
      $form['address_line2'] = [
        '#type' => 'textfield',
        '#title' => $this->t('Apt or unit #'),
        '#weight' => 5,
      ];
      $form['address_city'] = [
        '#type' => 'textfield',
        '#title' => $this->t('City or district'),
        // TODO add '#default_value' (for everything) so form repopulates after errors
        '#required' => TRUE,
        '#weight' => 6
      ];
      $form['address_zip'] = [
        '#type' => 'textfield',
        '#title' => $this->t('Postal code / ZIP'),
        '#required' => TRUE,
        '#weight' => 7
      ];
      $form['address_state'] = [
        '#type' => 'textfield',
        '#title' => $this->t('State or province'),
        '#required' => TRUE,
        '#weight' => 8
      ];
      $form['address_country'] = [
        '#type' => 'textfield',
        '#title' => $this->t('Country'),
        // @todo Default to the country of the site.
        '#default_value' => '',
        '#required' => TRUE,
        '#weight' => 9,
      ];
    }
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function actions(array $form, FormStateInterface $form_state) {
    $donation = $this->entity;
    $elements = parent::actions($form, $form_state);
    $elements['submit']['#value'] = $donation->getGiveForm()->getPaymentSubmitText();
    $elements['delete']['#title'] = $this->t('Cancel');
    return $elements;
  }

  /**
   * {@inheritdoc}
   */
  public function buildEntity(array $form, FormStateInterface $form_state) {
    /** @var \Drupal\give\DonationInterface $donation */
    $donation = parent::buildEntity($form, $form_state);
    $donation->setChangedTime(\Drupal::time()->getRequestTime());
    //$donation->amount->value = $form_state->getValue('amount'); // the original amount 800
    return $donation;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    /** @var \Drupal\give\Entity\Donation $donation */
    $donation = parent::validateForm($form, $form_state);
    if ($form_state->getValue('method') != PaymentMethod::GIVE_VIA_STRIPE) {
      return;
    }
    $give_config = $this->config('give.settings');
    $amount = (int)$donation->amount->value;
    $currency = $give_config->get('currency_symbol');
    // not sure how to do recurring payments.
    $this->giveStripe->setApiKey($give_config->get('stripe_secret_key'));
    // If the donation is recurring, we create a plan and a customer.
    if ($donation->isRecurring()) {
      // See https://stripe.com/docs/api/plans/object
      $plan_data = [
        "amount" => $amount, // cents
        "currency" => $currency,
        "interval" => $donation->getRecurrenceIntervalUnit(),
        "interval_count" => $donation->getRecurrenceIntervalCount(),
        "product" => [
          "name" => $donation->getRecurringProductName(),
          "statement_descriptor" => $donation->getStatementDescriptor(),
        ],
      ];
      try {
        $plan = $this->giveStripe->createPlan($plan_data);
      }
      catch (\Exception $e) {
        $this->processStripeErrors($e);
        $form_state->setErrorByName('stripe_errors', $e->getMessage());
        return;
      }
      // Create the customer with subscription plan on Stripe's servers - this
      // will charge the user's card.
      $customer_data = [
        "plan" => $plan->id,
        "source" => $donation->stripe_token->value,
        "email" => $donation->mail->value,
        "metadata" => [
          "give_form_id" => $donation->getGiveForm()->id(),
          "give_form_label" => $donation->getGiveForm()->label(),
          "email" => $donation->mail->value,
          "name" => $donation->getDonorName(),
        ],
      ];

      try {
        if ($this->giveStripe->createCustomer($customer_data)) {
          $this->entity->setCompleted();  // But not saved yet.
        }
      }
      catch (\Exception $e) {
        $form_state->setErrorByName('stripe_errors', $e->getMessage());
        $this->logger('give')->error('Stripe error: %msg.', ['%msg' => $e->getMessage()]);
      }
    }
    else {
      // Stripe Js has submitted the credit card details. which are not seen here.
      // Here we can change the amount if needed.
      try {
        $this->entity->setCardInfo($this->giveStripe);
        $this->entity->setCompleted();  // But not saved yet.
      }
      catch (\Exception $e) {
        $this->processStripeErrors($e);
        $form_state->setErrorByName('stripe_errors', $e->getMessage());
      }
    }
  }

  /**
   * Helper method to process stripe errors.
   *
   * There are three places they can happen, and we do a lot, so abstracting this out.
   */
  public function processStripeErrors(\Exception $e) {
    $this->logger('give')->error('Stripe error: %msg.', ['%msg' => $e->getMessage()]);
    \Drupal::messenger()->addError($e->getmessage());
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    $donation = $this->entity;
    $give_form = $donation->getGiveForm();
    $donation->save();

    if ($donation->method->value == 'card') {
      if ($donation->complete->value) {
        \Drupal::logger('stripe')->notice(
          'Payment of :amount processed in donation :donation',
          [':amount' => $donation->amount, ':donation' => $donation->toLink()->toString()]
        );
      }
      else {
        \Drupal::logger('stripe')->notice(
          'Failed to process :donation for :user',
          [':donation' => $donation->toLink(), ':user' => $donation->getOwner()->toLink()->toString()]
        );
        \Drupal::messenger(t(
          'Failed to process :donation for :user',
          [':donation' => $donation->toLink(), ':user' => $donation->getOwner()->toLink()->toString()])
        );
        return;
      }
    }
    $thanks_message = $this->token->replace(
      $give_form->get('thanks_message'),
      ['give_donation' => $donation]
    );
    $this->messenger()->addStatus($thanks_message);
    $url = Url::fromUserInput($give_form->getRedirectUri());
    $form_state->setRedirectUrl($url);

    if ($give_form->getRecipients()) {
      $this->mailHandler->sendDonationNotices($donation, User::load($this->currentUser()->id()));
    }
  }

  /**
   *
   * @param string $curr_code
   * @param int $cents
   * @param string $description
   * @return PaymentIntent
   */
  private function createPaymentIntent(string $curr_code, int $cents = 0, $description = '') : PaymentIntent {
    $this->giveStripe->setApiKey($this->config('give.settings')->get('stripe_secret_key'));
    $params = [
      'currency' => $curr_code,
      'payment_method_types' => ['card'],
      'description' => 'Membership for the rest of the year.'
    ];
    if ($cents) {
      $params['amount'] = $cents;
    }
    return $this->paymentIntent = \Stripe\PaymentIntent::create($params, []);
  }


}
